import logging
import os.path

from gi.repository import Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, gettext
from gaphor.ui.filedialog import open_file_dialog

log = logging.getLogger(__name__)


FILTERS = [
    (gettext("All Gaphor Models"), "*.gaphor", "application/x-gaphor"),
]


class AppFileManager(Service, ActionProvider):
    """Handle application level file loading."""

    def __init__(self, application):
        self.application = application
        self.last_dir = None

    def shutdown(self):
        pass

    @property
    def parent_window(self):
        return (
            self.application.active_session.get_service("main_window").window
            if self.application.active_session
            else None
        )

    @action(name="app.file-open", shortcut="<Primary>o")
    def action_open(self):
        """This menu action opens the standard model open dialog."""

        def open_files(filenames):
            for filename in filenames:
                if self.application.has_session(filename):
                    dialog = Gtk.MessageDialog(
                        message_type=Gtk.MessageType.QUESTION,
                        buttons=Gtk.ButtonsType.YES_NO,
                        text=gettext(
                            "{filename} is already opened. Do you want to switch to the opened window instead?"
                        ).format(filename=filename),
                    )
                    dialog.set_transient_for(self.parent_window)

                    def response(dialog, answer):
                        force_new_session = answer != Gtk.ResponseType.YES
                        dialog.destroy()
                        self.application.new_session(
                            filename=filename, force=force_new_session
                        )

                    dialog.connect("response", response)
                    dialog.set_modal(True)
                    dialog.show()
                else:
                    self.application.new_session(filename=filename)

                self.last_dir = os.path.dirname(filename)

        open_file_dialog(
            gettext("Open a Model"),
            open_files,
            parent=self.parent_window,
            dirname=self.last_dir,
            filters=FILTERS,
        )



"""Generalization --"""

from gaphor import UML
from gaphor.core.modeling import DrawContext
from gaphor.core.modeling.diagram import Diagram
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.shapes import Box, Text
from gaphor.diagram.support import represents
from gaphor.UML.recipes import stereotypes_str


@represents(
    UML.Generalization,
    head=UML.Generalization.specific,
    tail=UML.Generalization.general,
)
class GeneralizationItem(LinePresentation):
    def __init__(self, diagram: Diagram, id: str = None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                )
            ),
            style={"dash-style": ()},
        )

        self._handles[0].pos = (30, 20)
        self._handles[1].pos = (0, 0)

        self.watch("subject.appliedStereotype.classifier.name")

    def draw_tail(self, context: DrawContext):
        cr = context.cairo
        cr.line_to(15, 0)
        cr.move_to(15, -10)
        cr.line_to(0, 0)
        cr.line_to(15, 10)
        cr.close_path()



"""Service dedicated to exporting diagrams to a variety of file formats."""

import cairo
from gaphas.painter import BoundingBoxPainter, FreeHandPainter, PainterChain

from gaphor.core.modeling.diagram import StyledDiagram
from gaphor.diagram.painter import DiagramTypePainter, ItemPainter


def render(diagram, new_surface, padding=8):
    diagram.update_now(diagram.get_all_items())

    painter = new_painter(diagram)

    # Update bounding boxes with a temporary Cairo Context
    # (used for stuff like calculating font metrics)
    bounding_box = calc_bounding_box(diagram, painter)
    type_padding = diagram_type_height(diagram)

    w, h = (
        bounding_box.width + 2 * padding,
        bounding_box.height + 2 * padding + type_padding,
    )
    surface = new_surface(w, h)
    cr = cairo.Context(surface)

    bg_color = diagram.style(StyledDiagram(diagram)).get("background-color")
    if bg_color and bg_color[3]:
        cr.rectangle(0, 0, w, h)
        cr.set_source_rgba(*bg_color)
        cr.fill()

    cr.translate(-bounding_box.x + padding, -bounding_box.y + padding + type_padding)
    painter.paint(diagram.get_all_items(), cr)
    cr.show_page()
    return surface


def diagram_type_height(diagram):
    if not diagram.diagramType:
        return 0

    bounding_box = calc_bounding_box(diagram, DiagramTypePainter(diagram))

    return bounding_box.height


def calc_bounding_box(diagram, painter):
    tmpsurface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
    tmpcr = cairo.Context(tmpsurface)
    bounding_box = BoundingBoxPainter(painter).bounding_box(
        diagram.get_all_items(), tmpcr
    )
    tmpcr.show_page()
    tmpsurface.flush()
    return bounding_box


def save_svg(filename, diagram):
    surface = render(diagram, lambda w, h: cairo.SVGSurface(filename, w, h))
    surface.flush()
    surface.finish()


def save_png(filename, diagram):
    surface = render(
        diagram,
        lambda w, h: cairo.ImageSurface(cairo.FORMAT_ARGB32, int(w + 1), int(h + 1)),
    )
    surface.write_to_png(filename)


def save_pdf(filename, diagram):
    surface = render(diagram, lambda w, h: cairo.PDFSurface(filename, w, h))
    surface.flush()
    surface.finish()


def new_painter(diagram):
    style = diagram.style(StyledDiagram(diagram))
    sloppiness = style.get("line-style", 0.0)
    item_painter = (
        FreeHandPainter(ItemPainter(), sloppiness) if sloppiness else ItemPainter()
    )
    return PainterChain().append(item_painter).append(DiagramTypePainter(diagram))



from gaphas.connector import Handle, Port
from gaphas.guide import GuidedItemHandleMoveMixin
from gaphas.handlemove import HandleMove, ItemHandleMove
from gaphas.move import Move
from gaphas.types import Pos

from gaphor.diagram.connectors import Connector
from gaphor.diagram.presentation import connect
from gaphor.diagram.tools.grayout import GrayOutLineHandleMoveMixin
from gaphor.UML.actions.activity import ActivityItem, ActivityParameterNodeItem


@Connector.register(ActivityItem, ActivityParameterNodeItem)
class ActivityParameterNodeConnector:
    def __init__(
        self,
        activity: ActivityItem,
        parameter_node: ActivityParameterNodeItem,
    ) -> None:
        assert activity.diagram is parameter_node.diagram
        self.activity = activity
        self.parameter_node = parameter_node

    def allow(self, handle: Handle, port: Port) -> bool:
        return (
            bool(self.activity.diagram)
            and self.activity.diagram is self.parameter_node.diagram
            and self.parameter_node in self.activity.children
        )

    def connect(self, handle: Handle, port: Port) -> bool:
        return True

    def disconnect(self, handle: Handle) -> None:
        pass


@HandleMove.register(ActivityParameterNodeItem)
class ActivityParameterNodeItemHandleMove(
    GrayOutLineHandleMoveMixin, GuidedItemHandleMoveMixin, ItemHandleMove
):
    """We use a custom tool for moving parameter nodes.

    Parameter nodes always connect to their parent (Activity).
    """

    def connect(self, pos: Pos) -> None:
        item = self.item
        connect(item, self.handle, item.parent)
        self.view.model.request_update(item)


@Move.register(ActivityParameterNodeItem)
def activity_parameter_node_move(item, view):
    """We use a custom tool for moving parameter nodes.

    They can be moved while handle is connected and will automatically
    connect to a side of its parent.
    """
    return ActivityParameterNodeItemHandleMove(item, item.handles()[0], view)



import logging
import sys
import time

import cairo
import gi
from gi.repository import GLib, Gtk, Pango

from gaphor.abc import Service
from gaphor.application import Application, distribution

log = logging.getLogger(__name__)


class Status:
    def __init__(self, name):
        self.name = name
        self.status = "in progress"

    def complete(self):
        self.status = "complete"

    @property
    def completed(self):
        return self.status == "complete"

    def __repr__(self):
        return f"{self.name}: {self.status}"


def test(func):
    """A test function."""

    def wrapper(self):
        status = Status(func.__name__)
        self.statuses.append(status)
        return func(self, status)

    return wrapper


class SelfTest(Service):
    def __init__(self, application: Application):
        self.application = application
        self.statuses: list[Status] = []

    def shutdown(self):
        pass

    def init(self, gtk_app):
        windows_console_output_workaround()
        self.init_timer(gtk_app, timeout=20)
        self.test_library_versions()
        self.test_new_session()

    def init_timer(self, gtk_app, timeout):
        start = time.time()

        def callback():
            if all(status.completed for status in self.statuses):
                log.info(
                    "All tests have been completed in %.1fs %s",
                    time.time() - start,
                    self.statuses,
                )
                gtk_app.quit()
                return GLib.SOURCE_REMOVE
            elif time.time() > start + timeout:
                log.error("Not all tests have passed: %s", self.statuses)
                gtk_app.exit_code = 1
                gtk_app.quit()
                return GLib.SOURCE_REMOVE
            return GLib.SOURCE_CONTINUE

        GLib.timeout_add(priority=GLib.PRIORITY_LOW, interval=100, function=callback)

    @test
    def test_library_versions(self, status):
        log.info("Gaphor version:    %s", distribution().version)
        log.info("Python version:    %s", sys.version)
        log.info(
            "GTK version:       %d.%d.%d",
            Gtk.get_major_version(),
            Gtk.get_minor_version(),
            Gtk.get_micro_version(),
        )
        log.info("PyGObject version: %d.%d.%d", *gi.version_info)
        log.info("Pycairo version:   %s", cairo.version)
        log.info("Cairo version:     %s", cairo.cairo_version_string())
        log.info("Pango version:     %s", Pango.version_string())
        status.complete()

    @test
    def test_new_session(self, status):
        session = self.application.new_session()

        def check_new_session(session):
            main_window = session.get_service("main_window")

            if main_window.window and main_window.window.get_visible():
                status.complete()
                return GLib.SOURCE_REMOVE
            else:
                return GLib.SOURCE_CONTINUE

        GLib.idle_add(check_new_session, session, priority=GLib.PRIORITY_LOW)


def windows_console_output_workaround():
    if sys.platform == "win32":
        from gaphor.ui import LOG_FORMAT

        logging.basicConfig(
            level=logging.INFO,
            format=LOG_FORMAT,
            filename="gaphor-self-test.txt",
            filemode="w",
            force=True,
        )



"""Internationalization (i18n) support for Gaphor.

Translate text in to your native language using the gettext() function.
"""
__all__ = ["gettext"]

import functools
import gettext as _gettext
import importlib.resources
import logging
import xml.etree.ElementTree as etree

log = logging.getLogger(__name__)

try:
    with importlib.resources.path("gaphor", "__init__.py") as path:
        localedir = path.parent / "locale"
        translate = _gettext.translation("gaphor", localedir=localedir, fallback=True)
        gettext = translate.gettext

except OSError as e:
    log.warning(f"No translations were found: {e}")

    def gettext(s):
        return s


@functools.lru_cache(maxsize=None)
def translated_ui_string(package: str, ui_filename: str) -> str:
    with importlib.resources.path(package, ui_filename) as ui_path:
        ui_xml = etree.parse(ui_path)
    for node in ui_xml.findall(".//*[@translatable='yes']"):
        node.text = gettext(node.text) if node.text else ""
        del node.attrib["translatable"]
    return etree.tostring(ui_xml.getroot(), encoding="unicode", method="xml")



"""A generic way to handle errors in GUI applications.

This module also contains a ErrorHandlerAspect, which can be easily
attached to a class' method and will raise the error dialog when the
method exits with an exception.
"""

import pdb
import sys

from gi.repository import Gtk

from gaphor.i18n import gettext


def error_handler(message, secondary_message="", window=None):
    exc_type, exc_value, exc_traceback = sys.exc_info()

    dialog = Gtk.MessageDialog(message_type=Gtk.MessageType.ERROR, text=message)
    dialog.set_transient_for(window)

    if __debug__ and exc_traceback and sys.stdin.isatty():
        dialog.props.secondary_text = (
            f"{secondary_message}\n\n" if secondary_message else ""
        ) + gettext(
            "It looks like Gaphor is started from the command line. Do you want to open a debug session?"
        )
        dialog.add_buttons(gettext("Close"), 0, gettext("Start debug session"), 100)
    else:
        dialog.props.secondary_text = secondary_message
        dialog.add_button(gettext("Close"), 0)

    def response(dialog, answer):
        dialog.destroy()
        if exc_traceback and answer == 100:
            pdb.post_mortem(exc_traceback)

    dialog.connect("response", response)
    dialog.set_modal(True)
    dialog.show()
